"""Multidimensional minimizer based on the cross-maximum-volume principle.

This module contains the main function "ttopt" that finds the approximate
minimum of the given multidimensional array (tensor), which can represent a
discretized multivariable function.

Note:
    For the task of finding the extremum of a function of many variables or
    multidimensional array, a wrapper class "TTOpt" (from "ttopt.py") could be
    used. It provides a set of methods for discretizing the function, caching
    previously requested values and logging intermediate results. In this case,
    a wrapper "TTOpt.comp_min" should be passed to the function "ttopt" as its
    first argument (the method "TTOpt.minimize" provides the related interface).

"""
from maxvolpy.maxvol import maxvol
from maxvolpy.maxvol import rect_maxvol
import numpy as np


def ttopt(f, n, rmax=5, evals=None, Y0=None, fs_opt=None):
    """Find the minimum element of the implicitly given multidimensional array.

    This function computes the minimum of the implicitly given d-dimensional
    (d >= 2) array (tensor). The adaptive method based on the tensor train (TT)
    approximation and the cross-maximum-volume principle are used.

    Args:
        f (function): the function that returns tensor values for the given set
            of the indices. Its arguments are (I, i_min, y_min, opt_min), where
            "I" represents several multi-indices (samples) for calculation (it
            is 2D np.ndarray of the shape [samples, dimensions]), "i_min"
            represents the current multi-index of the argmin approximation (it
            is 1D np.ndarray of the shape [dimensions]; note that while the
            first call it will be None), "y_min" represents the current
            approximated minimum of the tensor (it is float; note that while
            the first call it will be None) and "opt_min" is the value of the
            auxiliary quantity corresponding to the multi-index "i_min" (it is
            used for debugging and in specific parallel calculations). The
            output of the function should be the corresponding values in the
            given indices (1D np.ndarray of the shape [samples]) and related
            values of the auxiliary quantities at the requested points (1D
            np.ndarray of the shape [samples] of any). If the function returns
            None instead of the tensor values, then the algorithm will be
            interrupted and the current approximation will be returned.
        n (list of len d of int): number of grid points for every dimension
            (i.e., the shape of the tensor). Note that the tensor must have a
            dimension of at least 2.
        rmax (int): maximum used rank for unfolding matrices.
        evals (int or float): number of available calls to function (i.e.,
            computational budget). If it is None, then the algorithm will run
            until the target function returns a None instead of the y-value.
        Y0 (list of 3D np.ndarrays): optional initial tensor in the TT-format
            (it should be represented as a list of the TT-cores). If it is not
            specified, then a random TT-tensor with TT-rank "rmax" will be used.
        fs_opt (float): the parameter of the smoothing function. If it is None,
            then "arctan" function will be used. Otherwise, the function
            "exp(-1 * fs_opt * (p - p0))" will be used.

    Returns:
        [np.ndarray, float]: the multi-index that gives the minimum value of the
        tensor (it is 1D np.ndarray of length "d" of int; i.e., "i_min") and
        the minimum value of the tensor (it is float; i.e., "y_min") that
        corresponds to the multi-index "i_min".

    """
    # Number of dimensions:
    d = len(n)

    # Number of possible function calls:
    evals = int(evals) if evals else None

    # Grid:
    J_grid = [np.reshape(np.arange(k), (-1, 1)) for k in n]

    # Initial tensor:
    Y0, R = ttopt_init(n, rmax, Y0, with_rank=True)

    # Selected multi-indices for all unfolding matrices:
    J = [None] * (d + 1)
    for i in range(d - 1):
        G = Y0[i].reshape(-1, Y0[i].shape[-1])
        q, r = np.linalg.qr(G)
        ind = _maxvol(q)
        J[i + 1] = _stack(J[i], J_grid[i], ind, l2r=True)

    i_min = None         # Approximation of argmin for tensor
    y_min = None         # Approximation of min for tensor (float('inf'))
    opt_min = None       # Additional option related to i_min

    eval = 0             # Number of performed calls to function:
    iter = 0             # Iteration (sweep) number
    i = d - 1            # Index of the current core (0, 1, ..., d-1)
    direction = -1       # Core traversal direction: right (+1) or left (-1)

    while True:
        # We select sample points [samples, d]:
        I = _merge(J[i], J[i+1], J_grid[i])

        # We check if the maximum number of requests has been exceeded:
        eval_curr = I.shape[0]
        if evals is not None and eval + eval_curr > evals:
            I = I[:(evals-eval), :]

        # We compute the function of interest "f" in the sample points I:
        y, opt = f(I, i_min, y_min, opt_min)

        # Function "f" can return None to interrupt the algorithm execution:
        if y is None:
            return i_min, y_min
        else:
            eval += y.size

        # We find and check the minimum value on a set of sampled points:
        i_min, y_min, opt_min = ttopt_find(I, y, opt, i_min, y_min, opt_min)

        # If the max number of requests exceeded, we interrupt the algorithm:
        if evals is not None and eval >= evals:
            return i_min, y_min

        # If computed points less then requested, we interrupt the algorithm:
        if y.shape[0] < I.shape[0]:
            return i_min, y_min

        # We transform sampled points into "core tensor" and smooth it out:
        Z = _reshape(y, (R[i], n[i], R[i + 1]))
        Z = ttopt_fs(Z, y_min, fs_opt)

        # We update ranks and points of interest according to the "core tensor":
        if direction > 0 and i < d - 1:
            # This is left to right sweep:
            Z = _reshape(Z, (R[i] * n[i], R[i + 1]))
            q, r = np.linalg.qr(Z)
            ind = _maxvol_rect(q)
            J[i + 1] = _stack(J[i], J_grid[i], ind, l2r=True)
            R[i + 1] = ind.size
        if direction < 0 and i > 0:
            # This is right to left sweep:
            Z = _reshape(Z, (R[i], n[i] * R[i + 1])).T
            q, r = np.linalg.qr(Z)
            R[i] = min(R[i], rmax)
            q = q[:, :R[i]]
            ind = _maxvol_rect(q)
            J[i] = _stack(J[i+1], J_grid[i], ind, l2r=False)
            R[i] = ind.size

        # We update the current core index according to the traversal direction
        # (when we go through the first or the latest core, then we inverse the
        # traversal direction and increment the iteration counter):
        i += direction
        if i == -1 or i == d:
            iter += 1
            direction = -direction
            i += direction

    return i_min, y_min


def ttopt_find(I, y, opt, i_min, y_min, opt_min):
    """Find the minimum value on a set of sampled points."""
    ind = np.argmin(y)
    y_min_curr = y[ind]

    if y_min is not None and y_min_curr >= y_min:
        return i_min, y_min, opt_min

    return I[ind, :], y_min_curr, opt[ind]


def ttopt_fs(p, p0=0., opt=None):
    """Smooth function that transforms max to min."""
    if opt is None:
        return np.pi/2 - np.arctan(p - p0)
    else:
        return np.exp(-1. * opt * (p - p0))


def ttopt_init(n, rmax, Y0=None, with_rank=False):
    """Build initial approximation for the main algorithm."""
    d = len(n)

    R = [1]
    for i in range(1, d):
        R.append(min(rmax, n[i-1] * R[i-1]))
    R.append(1)

    if Y0 is None:
        Y0 = [np.random.randn(R[i], n[i], R[i + 1]) for i in range(d)]

    if with_rank:
        return Y0, R
    else:
        return Y0


def _maxvol(A, tol=1.01, max_iters=100):
    return maxvol(A, tol=tol, max_iters=max_iters)[0]


def _maxvol_rect(A, kickrank=1, rf=1, tol=1.):
    if kickrank is not None and rf is not None:
        maxK = A.shape[1] + kickrank + rf
    else:
        maxK = None

    return rect_maxvol(A, tol=tol, min_add_K=kickrank, maxK=maxK,
        start_maxvol_iters=10, identity_submatrix=False)[0]


def _merge(J1, J2, Jg):
    n = Jg.shape[0]
    r1 = J1.shape[0] if J1 is not None else 1
    r2 = J2.shape[0] if J2 is not None else 1

    I = np.kron(np.kron(_ones(r2), Jg), _ones(r1))

    if J1 is not None:
        J1_ = np.kron(_ones(n * r2), J1)
        I = np.hstack((J1_, I))

    if J2 is not None:
        J2_ = np.kron(J2, _ones(r1 * n))
        I = np.hstack((I, J2_))

    return I


def _ones(k, m=1):
    return np.ones((k, m), dtype=int)


def _reshape(A, n):
    return np.reshape(A, n, order='F')


def _stack(J, Jg, ind, l2r=True):
    n = Jg.shape[0]
    r = J.shape[0] if J is not None else 1

    J_new = np.kron(Jg, _ones(r)) if l2r else np.kron(_ones(r), Jg)

    if J is not None:
        J_old = np.kron(_ones(n), J) if l2r else np.kron(J, _ones(n))
        J_new = np.hstack((J_old, J_new)) if l2r else np.hstack((J_new, J_old))

    return J_new[ind, :]



import numpy as np
from time import perf_counter as tpc


from .ttopt_raw import ttopt
from .ttopt_raw import ttopt_find
from .ttopt_raw_old import ttopt as ttopt_old


class TTOpt():
    """Multidimensional minimizer based on the cross-maximum-volume principle.

    Class for computation of the minimum for the implicitly given d-dimensional
    array (tensor) or a function of d-dimensional argument. An adaptive method
    based on the tensor train (TT) approximation and the cross-maximum volume
    principle is used. Cache of requested values (its usage leads to faster
    computation if one point is computed for a long time) and QTT-based
    representation of the grid (its usage in many cases leads to more accurate
    results) are supported.

    Args:
        f (function): the function of interest. Its argument X should represent
            several spatial points for calculation (is 2D numpy array of the
            shape [samples, d]) if "is_vect" flag is True, and it is one
            spatial point for calculation (is 1D numpy array of the shape [d])
            in the case if "is_vect" flag is False. For the case of the tensor
            approximation (if "is_func" flag is False), the argument X relates
            to the one or many (depending on the value of the flag "is_vect")
            multi-indices of the corresponding array/tensor. Function should
            return the values in the requested points (is 1D numpy array of the
            shape [samples] of float or only one float value depending on the
            value of "is_vect" flag). If "with_opt" flag is True, then function
            should also return the second argument (is 1D numpy array of the
            shape [samples] of any or just one value depending on the "is_vect"
            flag) which is the auxiliary quantity corresponding to the
            requested points (it is used for debugging and in specific parallel
            calculations; the value of this auxiliary quantity related to the
            "argmin" point will be passed to "callback" function).
        d (int): number of function dimensions.
        a (float or list of len d of float): grid lower bounds for every
            dimension. If a number is given, then this value will be used for
            each dimension.
        b (float or list of len d of float): grid upper bounds for every
            dimension. If a number is given, then this value will be used for
            each dimension.
        n (int or list of len d of int): number of grid points for every
            dimension. If a number is given, then this value will be used for
            each dimension. If this parameter is not specified, then instead of
            it the values for both "p" and "q" should be set.
        p (int): the grid size factor (if is given, then there will be n=p^q
            points for each dimension). This parameter can be specified instead
            of "n". If this parameter is specified, then the parameter "q" must
            also be specified, and in this case the QTT-based approach will be
            used.
        q (int): the grid size factor (if is given, then there will be n=p^q
            points for each dimension). This parameter can be specified instead
            of "n". If this parameter is specified, then the parameter "p" must
            also be specified, and in this case the QTT-based approach will be
            used.
        evals (int or float): the number of requests to the target function
            that will be made.
        name (str): optional display name for the function of interest. It is
            the empty string by default.
        callback (function): optional function that will be called after each
            optimization step (in Func.comp.min) with related info (it is used
            for debugging and in specific parallel calculations).
        x_min_real (list of len d): optional real value of x-minima (x). If
            this value is specified, then it will be used to display the
            current approximation error within the algorithm iterations (this
            is convenient for debugging and testing/research).
        y_min_real (float): optional real value of y-minima (y=f(x)). If
            this value is specified, then it will be used to display the
            current approximation error within the algorithm iterations (this
            is convenient for debugging and testing/research).
        is_func (bool): if flag is True, then we minimize the function (the
            arguments of f correspond to continuous spatial points), otherwise
            we approximate the tensor (the arguments of f correspond to
            discrete multidimensional tensor multi-indices). It is True by
            default.
        is_vect (bool): if flag is True, then function should accept 2D
            numpy array of the shape [samples, d] (batch of points or indices)
            and return 1D numpy array of the shape [samples]. Otherwise, the
            function should accept 1D numpy array (one multidimensional point)
            and return the float value. It is True by default.
        with_cache (bool): if flag is True, then all requested values are
            stored and retrieved from the storage upon repeated requests.
            Note that this leads to faster computation if one point is
            computed for a long time. On the other hand, this can in some
            cases slow down the process, due to the additional time spent
            on checking the storage and using unvectorized slow loops in
            python. It is False by default.
        with_log (bool): if flag is True, then text messages will be
            displayed during the optimizer query process. It is False by
            default.
        with_opt (bool): if flag is True, then function of interest returns
            opts related to output y (scalar or vector) as second argument
            (it will be also saved and passed to "callback" function). It is
            False by default.
        use_old (bool): if flag is True, then old TTOpt algorithm will be used.
            This is relevant for the possibility of reproducing old
            calculations reflected in publications.
        with_full_info (bool): if flag is True, then the full information will
            be saved, including multi-indices of requested points (it is used
            by animation function) and best found multi-indices and points.
            Note that the inclusion of this flag can significantly slow down
            the process of the algorithm. It is False by default.
        with_wrn (bool): if flag is True, then warning messages will be
            presented (in the current version, it can only be messages about
            early convergence when using the cache). It is True by default.

    Note:
        Call "calc" to evaluate function for one tensor multi-index and call
        "comp" to evaluate function in the set of multi-indices (both of these
        functions can be called regardless of the value of the flag "is_vect").
        Call "minimize" to find the global minimum of the function of interest
        by the TTOpt-algorithm.

    """

    def __init__(self, f, d, a=None, b=None, n=None, p=None, q=None, evals=None, name=None, callback=None, x_min_real=None, y_min_real=None, is_func=True, is_vect=True, with_cache=False, with_log=False, with_opt=False, use_old=False, with_full_info=False, with_wrn=True):
        # Set the target function and its dimension:
        self.f = f
        self.d = int(d)

        # Set grid lower bound:
        if isinstance(a, (int, float)):
            self.a = np.ones(self.d, dtype=float) * a
        elif a is not None:
            self.a = np.asanyarray(a, dtype=float)
        else:
            if is_func:
                raise ValueError('Grid lower bound (a) should be set')
            self.a = None
        if self.a is not None and self.a.size != self.d:
            raise ValueError('Grid lower bound (a) has invalid shape')

        # Set grid upper bound:
        if isinstance(b, (int, float)):
            self.b = np.ones(self.d, dtype=float) * b
        elif b is not None:
            self.b = np.asanyarray(b, dtype=float)
        else:
            if is_func:
                raise ValueError('Grid upper bound (b) should be set')
            self.b = None
        if self.b is not None and self.b.size != self.d:
            raise ValueError('Grid upper bound (b) has invalid shape')

        # Set number of grid points:
        if n is None:
            if p is None or q is None:
                raise ValueError('If n is not set, then p and q should be set')
            self.p = int(p)
            self.q = int(q)
            self.n = np.ones(self.d * self.q, dtype=int) * self.p
            self.n_func = np.ones(self.d, dtype=int) * (self.p**self.q)
        else:
            if p is not None or q is not None:
                raise ValueError('If n is set, then p and q should be None')
            self.p = None
            self.q = None
            if isinstance(n, (int, float)):
                self.n = np.ones(self.d, dtype=int) * int(n)
            else:
                self.n = np.asanyarray(n, dtype=int)
            self.n_func = self.n.copy()
        if self.n_func.size != self.d:
            raise ValueError('Grid size (n/p/q) has invalid shape')

        # Set other options according to the input arguments:
        self.evals = int(evals) if evals else None
        self.name = name or ''
        self.callback = callback
        self.x_min_real = x_min_real
        self.y_min_real = y_min_real
        self.is_func = bool(is_func)
        self.is_vect = bool(is_vect)
        self.with_cache = bool(with_cache)
        self.with_log = bool(with_log)
        self.with_opt = bool(with_opt)
        self.use_old = bool(use_old)
        self.with_full_info = bool(with_full_info)
        self.with_wrn = bool(with_wrn)

        # Inner variables:
        self.cache = {}     # Cache for the results of requests to function
        self.cache_opt = {} # Cache for the options while requests to function
        self.k_cache = 0    # Number of requests, then cache was used
        self.k_cache_curr = 0
        self.k_evals = 0    # Number of requests, then function was called
        self.k_evals_curr = 0
        self.t_evals = 0.   # Total time of function calls
        self.t_total = 0.   # Total time of computations (including cache usage)
        self.t_minim = 0    # Total time of work for minimizator
        self._opt = None    # Function opts related to its output

        # Current minima:
        self.i_min = None
        self.x_min = None

        # Approximations for argmin/min/opts of the function while iterations:
        self.I_list = []
        self.i_min_list = []
        self.x_min_list = []
        self.y_min_list = []
        self.opt_min_list = []
        self.evals_min_list = []
        self.cache_min_list = []

    @property
    def e_x(self):
        """Current error for approximation of argmin of the function."""
        if self.x_min_real is not None and self.x_min is not None:
            return np.linalg.norm(self.x_min - self.x_min_real)

    @property
    def e_y(self):
        """Current error for approximation of the minumum of the function."""
        if self.y_min_real is not None and self.y_min is not None:
            return np.abs(self.y_min - self.y_min_real)

    @property
    def k_total(self):
        """Total number of requests (both function calls and cache usage)."""
        return self.k_cache + self.k_evals

    @property
    def opt_min(self):
        """Current value of option of the function related to min-point."""
        return self.opt_min_list[-1] if len(self.opt_min_list) else None

    @property
    def t_evals_mean(self):
        """Average time spent to real function call for 1 point."""
        return self.t_evals / self.k_evals if self.k_evals else 0.

    @property
    def t_total_mean(self):
        """Average time spent to return one function value."""
        return self.t_total / self.k_total if self.k_total else 0.

    @property
    def y_min(self):
        """Current approximation of min of the function of interest."""
        return self.y_min_list[-1] if len(self.y_min_list) else None

    def calc(self, i):
        """Calculate the function for the given multiindex.

        Args:
            i (np.ndarray): the input for the function, that is 1D numpy array
                of the shape [d] of int (indices).

        Returns:
            float: the output of the function.

        """
        self.k_cache_curr = 0

        if self.is_vect:
            return self.comp(i.reshape(1, -1))[0]

        t_total = tpc()

        if not self.with_cache:
            y = self._eval(i)
            self.t_total += tpc() - t_total
            return y

        s = self.i2s(i.astype(int).tolist())

        if not s in self.cache:
            y = self._eval(i)
            self.cache[s] = y
            self.cache_opt[s] = self._opt
        else:
            y = self.cache[s]
            self._opt = self.cache_opt[s]
            self.k_cache_curr = 1
            self.k_cache += self.k_cache_curr

        self.t_total += tpc() - t_total

        return y

    def comp(self, I):
        """Compute the function for the set of multi-indices (samples).

        Args:
            I (np.ndarray): the inputs for the function, that are collected in
                2D numpy array of the shape [samples, d] of int (indices).

        Returns:
            np.ndarray: the outputs of the function, that are collected in
            1D numpy array of the shape [samples].

        Note:
            The set of points (I) should not contain duplicate points. If it
            contains duplicate points (that are not in the cache), then each of
            them will be recalculated without using the cache.

        """
        self.k_cache_curr = 0

        if not self.is_vect:
            Y, _opt = [], []
            for i in I:
                Y.append(self.calc(i))
                _opt.append(self._opt)
            self._opt = _opt
            return np.array(Y)

        t_total = tpc()

        if not self.with_cache:
            Y = self._eval(I)
            self.t_total += tpc() - t_total
            return Y

        # Requested points:
        I = I.tolist()

        # Points that are not presented in the cache:
        J = [i for i in I if self.i2s(i) not in self.cache]
        self.k_cache_curr = len(I) - len(J)
        self.k_cache += self.k_cache_curr

        # We add new points (J) to the storage:
        if len(J):
            Z = self._eval(J)
            for k, j in enumerate(J):
                s = self.i2s(j)
                self.cache[s] = Z[k]
                self.cache_opt[s] = self._opt[k]

        # We obtain the values for requested points from the updated storage:
        Y = np.array([self.cache[self.i2s(i)] for i in I])
        self._opt = np.array([self.cache_opt[self.i2s(i)] for i in I])

        self.t_total += tpc() - t_total

        return Y

    def comp_min(self, I, i_min=None, y_min=None, opt_min=None):
        """Compute the function for the set of points and save current minimum.

        This helper function (this is wrapper for function "comp") can be
        passed to the optimizer. When making requests, the optimizer must pass
        the grid points of interest (I) as arguments, as well as the current
        approximation of the argmin (i_min), the corresponding value (y_min)
        and related option value (opt_min).

        """
        # We return None if the limit for function requests is exceeded:
        if self.evals is not None and self.k_evals >= self.evals:
            return None, None

        # We return None if the number of requests to the cache is 2 times
        # higher than the number of requests to the function:
        if self.with_cache:
            if self.k_cache >= self.evals and self.k_cache >= 2 * self.k_evals:
                text = '!!! TTOpt warning : '
                text += 'the number of requests to the cache is 2 times higher '
                text += 'than the number of requests to the function. '
                text += 'The work is finished before max func-evals reached.'
                if self.with_wrn:
                    print(text)
                return None, None

        # We truncate the list of requested points if it exceeds the limit:
        eval_curr = I.shape[0]
        is_last = self.evals is not None and self.k_evals+eval_curr>=self.evals
        if is_last:
            I = I[:(self.evals-self.k_evals), :]

        if self.q:
            # The QTT is used, hence we should transform the indices:
            if I is not None:
                I = self.qtt_parse_many(I)
            if i_min is not None:
                i_min = self.qtt_parse_many(i_min.reshape(1, -1))[0, :]

        Y = self.comp(I)

        # If this is last iteration, we should "manually" check for y_opt_new:
        if is_last:
            i_min, y_min, opt_min = ttopt_find(
                I, Y, self._opt, i_min, y_min, opt_min)

        if i_min is None:
            return Y, self._opt

        if self.is_func:
            x_min = self.i2x(i_min)
        else:
            x_min = i_min.copy()

        self.i_min = i_min.copy()
        self.x_min = x_min.copy()

        self.y_min_list.append(y_min)
        self.opt_min_list.append(opt_min)
        self.evals_min_list.append(self.k_evals_curr)
        self.cache_min_list.append(self.k_cache_curr)

        if self.with_full_info:
            self.I_list.append(I)
            self.i_min_list.append(self.i_min.copy())
            self.x_min_list.append(self.x_min.copy())

        is_better = len(self.y_min_list) == 1 or (y_min < self.y_min_list[-2])
        if self.callback and is_better:
            last = {'last': [x_min, y_min, i_min, opt_min, self.k_evals]}
            self.callback(last)

        if self.with_log:
            print(self.info(is_final=False))

        return Y, self._opt

    def i2s(self, i):
        """Transform array of int like [1, 2, 3] into string like '1-2-3'."""
        return '-'.join([str(v) for v in i])

    def i2x(self, i):
        """Transform multiindex into point of the uniform grid."""
        t = i * 1. / (self.n_func - 1)
        x = t * (self.b - self.a) + self.a
        return x

    def i2x_many(self, I):
        """Transform multiindices (samples) into grid points."""
        A = np.repeat(self.a.reshape((1, -1)), I.shape[0], axis=0)
        B = np.repeat(self.b.reshape((1, -1)), I.shape[0], axis=0)
        N = np.repeat(self.n_func.reshape((1, -1)), I.shape[0], axis=0)
        T = I * 1. / (N - 1)
        X = T * (B - A) + A
        return X

    def info(self, with_e_x=True, with_e_y=True, is_final=True):
        """Return text description of the progress of optimizer work."""
        text = ''

        if self.name:
            name = self.name + f'-{self.d}d'
            name += ' ' * max(0, 10 - len(name))
            text += name + ' | '

        if self.with_cache:
            text += f'evals={self.k_evals:-8.2e}+{self.k_cache:-8.2e} | '
        else:
            text += f'evals={self.k_total:-8.2e} | '

        if is_final:
            text += f't_all={self.t_minim:-8.2e} | '
        else:
            text += f't_cur={self.t_total:-8.2e} | '

        if self.y_min_real is None and self.y_min is not None:
            text += f'y={self.y_min:-13.6e} '
        else:
            if with_e_x and self.e_x is not None:
                text += f'e_x={self.e_x:-8.2e} '
            if with_e_y and self.e_y is not None:
                text += f'e_y={self.e_y:-8.2e} '

        return text

    def minimize(self, rmax=10, Y0=None, fs_opt=1., add_opt_inner=True, add_opt_outer=False, add_opt_rect=False, add_rnd_inner=False, add_rnd_outer=False, J0=None):
        """Perform the function minimization process by TT-based approach.

        Args:
            rmax (int): maximum TT-rank.
            Y0 (list of 3D np.ndarrays of float): optional initial tensor in
                the TT format as a list of the TT-cores.
            fs_opt (float): the parameter of the smoothing function. If it is
                None, then "arctan" function will be used. Otherwise, the
                function "exp(-1 * fs_opt * (p - p0))" will be used.

        """
        t_minim = tpc()

        if self.use_old:
            i_min, y_min = ttopt_old(self.comp_min, self.n, rmax, None, Y0,
                fs_opt)
        else:
            i_min, y_min = ttopt(self.comp_min, self.n, rmax, None, Y0,
                fs_opt, add_opt_inner, add_opt_outer, add_opt_rect,
                add_rnd_inner, add_rnd_outer, J0)

        self.t_minim = tpc() - t_minim

    def qtt_parse_many(self, I_qtt):
        """Transform tensor indices from QTT (long) to base (short) format."""
        samples = I_qtt.shape[0]
        n_qtt = [self.n[0]]*self.q
        I = np.zeros((samples, self.d))
        for i in range(self.d):
            J_curr = I_qtt[:, self.q*i:self.q*(i+1)].T
            I[:, i] = np.ravel_multi_index(J_curr, n_qtt, order='F')
        return I

    def s2i(self, s):
        """Transforms string like '1-2-3' into array of int like [1, 2, 3]."""
        return np.array([int(v) for v in s.split('-')], dtype=int)

    def _eval(self, i):
        """Helper that computes target function in one or many points."""
        t_evals = tpc()

        i = np.asanyarray(i, dtype=int)
        is_many = len(i.shape) == 2

        if self.is_func:
            x = self.i2x_many(i) if is_many else self.i2x(i)
        else:
            x = i

        if self.with_opt:
            y, self._opt = self.f(x)
        else:
            y = self.f(x)
            self._opt = [None for _ in range(y.size)] if is_many else None

        self.k_evals_curr = y.size if is_many else 1
        self.k_evals += self.k_evals_curr
        self.t_evals += tpc() - t_evals

        return y



"""The demo of using ttopt. Simple example with "scalar" input.

We'll find the minimum for the simple analytic function of the form f(x), where
input x is the [dimension] numpy array (not vectorized!).

Run it from the root of the project as "python demo/vect.py".

As a result of the script work we expect the output in console like this:
"
...
Simple-5d | k=1.00e+04 | t_cur=1.54e-01 | e_x=3.55e-02 e_y=1.03e-04
----------------------------------------------------------------------
Simple-5d | k=1.00e+04 | t_all=1.75e-01 | e_x=3.55e-02 e_y=1.03e-04
"

"""
import numpy as np
from scipy.optimize import rosen


from ttopt import TTOpt
from ttopt import ttopt_init


np.random.seed(42)


d = 5                       # Number of function dimensions:
rmax = 4                    # Maximum TT-rank while cross-like iterations
def f(x):                   # Target function
    return np.sin(0.1 * x[0])**2 + 0.1 * np.sum(x[1:]**2)


# We initialize the TTOpt class instance with the correct parameters:
tto = TTOpt(
    f=f,                    # Function for minimization. X is [samples, dim]
    d=d,                    # Number of function dimensions
    a=-1.,                  # Grid lower bound (number or list of len d)
    b=+1.,                  # Grid upper bound (number or list of len d)
    n=2**6,                 # Number of grid points (number or list of len d)
    evals=1.E+4,            # Number of function evaluations
    name='Simple',          # Function name for log (this is optional)
    x_min_real=np.zeros(d), # Real value of x-minima (x; this is for test)
    y_min_real=0.,          # Real value of y-minima (y=f(x); this is for test)
    is_vect=False,          # The function accepts only one spatial point
    with_log=True)


# And now we launching the minimizer:
tto.minimize(rmax)


# We can extract the results of the computation:
x = tto.x_min          # The found value of the minimum of the function (x)
y = tto.y_min          # The found value of the minimum of the function (y=f(x))
k_c = tto.k_cache      # Total number of cache usage (should be 0 in this demo)
k_e = tto.k_evals      # Total number of requests to func (is always = evals)
k_t = tto.k_total      # Total number of requests (k_cache + k_evals)
t_f = tto.t_evals_mean # Average time spent to real function call for 1 point
                       # ... (see "ttopt.py" and docs for more details)


# We log the final state:
print('-' * 70 + '\n' + tto.info() +'\n\n')



"""The demo of using ttopt. Example for tensor minimization.

We'll find the minimum for the given simple d-dimensional tensor with elements:
Y[i_1, i_2, ..., i_d] = (i_1 - 2)^2 + (i_2 - 3)^2 + i_2^4 + i_3^4 + ... + i_d^4.

Run it from the root of the project as "python demo/tensor.py".

As a result of the script work we expect the output in console like this:
"
...
Tensor-10d | k=1.00e+05 | t_cur=9.04e-03 | e_x=0.00e+00 e_y=0.00e+00
----------------------------------------------------------------------
Tensor-10d | k=1.00e+05 | t_all=1.21e-01 | e_x=0.00e+00 e_y=0.00e+00
y_opt :  0
i_opt :  [2 3 0 0 0 0 0 0 0 0]
"

"""
import numpy as np


from ttopt import TTOpt
from ttopt import ttopt_init


np.random.seed(42)


d = 10                      # Number of function dimensions
p = 2
q = 10
n = p**q                    # Mode size for the tensor
rmax = 2                    # Maximum TT-rank while cross-like iterations
def f(I):                   # Target function (return tensor element)
    return (I[:, 0] - 2)**2 + (I[:, 1] - 3)**2 + np.sum(I[:, 2:]**4, axis=1)


# Real value of x-minima:
x_min_real = np.zeros(d)
x_min_real[0] = 2
x_min_real[1] = 3


# We initialize the TTOpt class instance with the correct parameters:
tto = TTOpt(
    f=f,                    # Function for minimization. X is [samples, dim]
    d=d,                    # Number of function dimensions
    n=n,                    # Number of grid points (number or list of len d)
    evals=1.E+5,            # Number of function evaluations
    name='Tensor',          # Function name for log (this is optional)
    x_min_real=x_min_real,  # Real value of x-minima (x; this is for test)
    y_min_real=0.,          # Real value of y-minima (y=f(x); this is for test)
    is_func=False,          # We approximate the tensor (not a function)
    with_log=True)


# And now we launching the minimizer:
tto.minimize(rmax)


# We can extract the results of the computation:
i = tto.i_min          # The found value of the minimum (multi-index)
y = tto.y_min          # The found value of the minimum of the function (y=f(x))
k_c = tto.k_cache      # Total number of cache usage (should be 0 in this demo)
k_e = tto.k_evals      # Total number of requests to func (is always = evals)
k_t = tto.k_total      # Total number of requests (k_cache + k_evals)
t_f = tto.t_evals_mean # Average time spent to real function call for 1 point
                       # ... (see "ttopt.py" and docs for more details)


# We log the final state:
print('-' * 70 + '\n' + tto.info())
print('y_opt : ', y)
print('i_opt : ', i)



"""Implementation of various Evolution Strategies.

This script was downloaded from the repo https://github.com/hardmaru/estool.
The package "cma" (https://github.com/CMA-ES/pycma) must be installed for it to
work correctly ("pip install cma").

"""
import numpy as np


def compute_ranks(x):
  """
  Returns ranks in [0, len(x))
  Note: This is different from scipy.stats.rankdata, which returns ranks in [1, len(x)].
  (https://github.com/openai/evolution-strategies-starter/blob/master/es_distributed/es.py)
  """
  assert x.ndim == 1
  ranks = np.empty(len(x), dtype=int)
  ranks[x.argsort()] = np.arange(len(x))
  return ranks

def compute_centered_ranks(x):
  """
  https://github.com/openai/evolution-strategies-starter/blob/master/es_distributed/es.py
  """
  y = compute_ranks(x.ravel()).reshape(x.shape).astype(np.float32)
  y /= (x.size - 1)
  y -= .5
  return y

def compute_weight_decay(weight_decay, model_param_list):
  model_param_grid = np.array(model_param_list)
  return - weight_decay * np.mean(model_param_grid * model_param_grid, axis=1)

# adopted from:
# https://github.com/openai/evolution-strategies-starter/blob/master/es_distributed/optimizers.py

class Optimizer(object):
  def __init__(self, pi, epsilon=1e-08):
    self.pi = pi
    self.dim = pi.num_params
    self.epsilon = epsilon
    self.t = 0

  def update(self, globalg):
    self.t += 1
    step = self._compute_step(globalg)
    theta = self.pi.mu
    ratio = np.linalg.norm(step) / (np.linalg.norm(theta) + self.epsilon)
    self.pi.mu = theta + step
    return ratio

  def _compute_step(self, globalg):
    raise NotImplementedError


class BasicSGD(Optimizer):
  def __init__(self, pi, stepsize):
    Optimizer.__init__(self, pi)
    self.stepsize = stepsize

  def _compute_step(self, globalg):
    step = -self.stepsize * globalg
    return step

class SGD(Optimizer):
  def __init__(self, pi, stepsize, momentum=0.9):
    Optimizer.__init__(self, pi)
    self.v = np.zeros(self.dim, dtype=np.float32)
    self.stepsize, self.momentum = stepsize, momentum

  def _compute_step(self, globalg):
    self.v = self.momentum * self.v + (1. - self.momentum) * globalg
    step = -self.stepsize * self.v
    return step


class Adam(Optimizer):
  def __init__(self, pi, stepsize, beta1=0.99, beta2=0.999):
    Optimizer.__init__(self, pi)
    self.stepsize = stepsize
    self.beta1 = beta1
    self.beta2 = beta2
    self.m = np.zeros(self.dim, dtype=np.float32)
    self.v = np.zeros(self.dim, dtype=np.float32)

  def _compute_step(self, globalg):
    a = self.stepsize * np.sqrt(1 - self.beta2 ** self.t) / (1 - self.beta1 ** self.t)
    self.m = self.beta1 * self.m + (1 - self.beta1) * globalg
    self.v = self.beta2 * self.v + (1 - self.beta2) * (globalg * globalg)
    step = -a * self.m / (np.sqrt(self.v) + self.epsilon)
    return step

class CMAES:
  '''CMA-ES wrapper.'''
  def __init__(self, num_params,      # number of model parameters
               sigma_init=0.10,       # initial standard deviation
               popsize=255,           # population size
               weight_decay=0.01, # weight decay coefficient
               seed=0,
               x0=None):

    self.num_params = num_params
    self.sigma_init = sigma_init
    self.popsize = popsize
    self.weight_decay = weight_decay
    self.solutions = None
    self.seed=seed

    import cma
    if x0 is None:
        x0 = self.num_params * [0.]
    self.es = cma.CMAEvolutionStrategy( x0,
                                        self.sigma_init,
                                        {'seed':self.seed,'popsize': self.popsize}
                                      )

  def rms_stdev(self):
    sigma = self.es.result[6]
    return np.mean(np.sqrt(sigma*sigma))

  def ask(self):
    '''returns a list of parameters'''
    self.solutions = np.array(self.es.ask())
    return self.solutions

  def tell(self, reward_table_result):
    reward_table = -np.array(reward_table_result)
    if self.weight_decay > 0:
      l2_decay = compute_weight_decay(self.weight_decay, self.solutions)
      reward_table += l2_decay
    self.es.tell(self.solutions, (reward_table).tolist()) # convert minimizer to maximizer.

  def current_param(self):
    return self.es.result[5] # mean solution, presumably better with noise

  def set_mu(self, mu):
    pass

  def best_param(self):
    return self.es.result[0] # best evaluated solution

  def result(self): # return best params so far, along with historically best reward, curr reward, sigma
    r = self.es.result
    return (r[0], -r[1], -r[1], r[6])

class SimpleGA:
  '''Simple Genetic Algorithm.'''
  def __init__(self, num_params,      # number of model parameters
               sigma_init=0.1,        # initial standard deviation
               sigma_decay=0.999,     # anneal standard deviation
               sigma_limit=0.01,      # stop annealing if less than this
               popsize=256,           # population size
               elite_ratio=0.1,       # percentage of the elites
               forget_best=False,     # forget the historical best elites
               weight_decay=0.01,     # weight decay coefficient
              ):

    self.num_params = num_params
    self.sigma_init = sigma_init
    self.sigma_decay = sigma_decay
    self.sigma_limit = sigma_limit
    self.popsize = popsize

    self.elite_ratio = elite_ratio
    self.elite_popsize = int(self.popsize * self.elite_ratio)

    self.sigma = self.sigma_init
    self.elite_params = np.zeros((self.elite_popsize, self.num_params))
    self.elite_rewards = np.zeros(self.elite_popsize)
    self.best_param = np.zeros(self.num_params)
    self.best_reward = 0
    self.first_iteration = True
    self.forget_best = forget_best
    self.weight_decay = weight_decay

  def rms_stdev(self):
    return self.sigma # same sigma for all parameters.

  def ask(self):
    '''returns a list of parameters'''
    self.epsilon = np.random.randn(self.popsize, self.num_params) * self.sigma
    solutions = []

    def mate(a, b):
      c = np.copy(a)
      idx = np.where(np.random.rand((c.size)) > 0.5)
      c[idx] = b[idx]
      return c

    elite_range = range(self.elite_popsize)
    for i in range(self.popsize):
      idx_a = np.random.choice(elite_range)
      idx_b = np.random.choice(elite_range)
      child_params = mate(self.elite_params[idx_a], self.elite_params[idx_b])
      solutions.append(child_params + self.epsilon[i])

    solutions = np.array(solutions)
    self.solutions = solutions

    return solutions

  def tell(self, reward_table_result):
    # input must be a numpy float array
    assert(len(reward_table_result) == self.popsize), "Inconsistent reward_table size reported."

    reward_table = np.array(reward_table_result)

    if self.weight_decay > 0:
      l2_decay = compute_weight_decay(self.weight_decay, self.solutions)
      reward_table += l2_decay

    if self.forget_best or self.first_iteration:
      reward = reward_table
      solution = self.solutions
    else:
      reward = np.concatenate([reward_table, self.elite_rewards])
      solution = np.concatenate([self.solutions, self.elite_params])

    idx = np.argsort(reward)[::-1][0:self.elite_popsize]

    self.elite_rewards = reward[idx]
    self.elite_params = solution[idx]

    self.curr_best_reward = self.elite_rewards[0]

    if self.first_iteration or (self.curr_best_reward > self.best_reward):
      self.first_iteration = False
      self.best_reward = self.elite_rewards[0]
      self.best_param = np.copy(self.elite_params[0])

    if (self.sigma > self.sigma_limit):
      self.sigma *= self.sigma_decay

  def current_param(self):
    return self.elite_params[0]

  def set_mu(self, mu):
    pass

  def best_param(self):
    return self.best_param

  def result(self): # return best params so far, along with historically best reward, curr reward, sigma
    return (self.best_param, self.best_reward, self.curr_best_reward, self.sigma)

class OpenES:
  ''' Basic Version of OpenAI Evolution Strategies.'''
  def __init__(self, num_params,             # number of model parameters
               sigma_init=0.1,               # initial standard deviation
               sigma_decay=0.999,            # anneal standard deviation
               sigma_limit=0.01,             # stop annealing if less than this
               learning_rate=0.01,           # learning rate for standard deviation
               learning_rate_decay = 0.9999, # annealing the learning rate
               learning_rate_limit = 0.001,  # stop annealing learning rate
               popsize=256,                  # population size
               antithetic=False,             # whether to use antithetic sampling
               weight_decay=0.01,            # weight decay coefficient
               rank_fitness=True,            # use rank rather than fitness numbers
               forget_best=True):            # forget historical best

    self.num_params = num_params
    self.sigma_decay = sigma_decay
    self.sigma = sigma_init
    self.sigma_init = sigma_init
    self.sigma_limit = sigma_limit
    self.learning_rate = learning_rate
    self.learning_rate_decay = learning_rate_decay
    self.learning_rate_limit = learning_rate_limit
    self.popsize = popsize
    self.antithetic = antithetic
    if self.antithetic:
      assert (self.popsize % 2 == 0), "Population size must be even"
      self.half_popsize = int(self.popsize / 2)

    self.reward = np.zeros(self.popsize)
    self.mu = np.zeros(self.num_params)
    self.best_mu = np.zeros(self.num_params)
    self.best_reward = 0
    self.first_interation = True
    self.forget_best = forget_best
    self.weight_decay = weight_decay
    self.rank_fitness = rank_fitness
    if self.rank_fitness:
      self.forget_best = True # always forget the best one if we rank
    # choose optimizer
    self.optimizer = Adam(self, learning_rate)

  def rms_stdev(self):
    sigma = self.sigma
    return np.mean(np.sqrt(sigma*sigma))

  def ask(self):
    '''returns a list of parameters'''
    # antithetic sampling
    if self.antithetic:
      self.epsilon_half = np.random.randn(self.half_popsize, self.num_params)
      self.epsilon = np.concatenate([self.epsilon_half, - self.epsilon_half])
    else:
      self.epsilon = np.random.randn(self.popsize, self.num_params)

    self.solutions = self.mu.reshape(1, self.num_params) + self.epsilon * self.sigma

    return self.solutions

  def tell(self, reward_table_result):
    # input must be a numpy float array
    assert(len(reward_table_result) == self.popsize), "Inconsistent reward_table size reported."

    reward = np.array(reward_table_result)

    if self.rank_fitness:
      reward = compute_centered_ranks(reward)

    if self.weight_decay > 0:
      l2_decay = compute_weight_decay(self.weight_decay, self.solutions)
      reward += l2_decay

    idx = np.argsort(reward)[::-1]

    best_reward = reward[idx[0]]
    best_mu = self.solutions[idx[0]]

    self.curr_best_reward = best_reward
    self.curr_best_mu = best_mu

    if self.first_interation:
      self.first_interation = False
      self.best_reward = self.curr_best_reward
      self.best_mu = best_mu
    else:
      if self.forget_best or (self.curr_best_reward > self.best_reward):
        self.best_mu = best_mu
        self.best_reward = self.curr_best_reward

    # main bit:
    # standardize the rewards to have a gaussian distribution
    normalized_reward = (reward - np.mean(reward)) / np.std(reward)
    change_mu = 1./(self.popsize*self.sigma)*np.dot(self.epsilon.T, normalized_reward)

    #self.mu += self.learning_rate * change_mu

    self.optimizer.stepsize = self.learning_rate
    update_ratio = self.optimizer.update(-change_mu)

    # adjust sigma according to the adaptive sigma calculation
    if (self.sigma > self.sigma_limit):
      self.sigma *= self.sigma_decay

    if (self.learning_rate > self.learning_rate_limit):
      self.learning_rate *= self.learning_rate_decay

  def current_param(self):
    return self.curr_best_mu

  def set_mu(self, mu):
    self.mu = np.array(mu)

  def best_param(self):
    return self.best_mu

  def result(self): # return best params so far, along with historically best reward, curr reward, sigma
    return (self.best_mu, self.best_reward, self.curr_best_reward, self.sigma)

class PEPG:
  '''Extension of PEPG with bells and whistles.'''
  def __init__(self, num_params,             # number of model parameters
               sigma_init=0.10,              # initial standard deviation
               sigma_alpha=0.20,             # learning rate for standard deviation
               sigma_decay=0.999,            # anneal standard deviation
               sigma_limit=0.01,             # stop annealing if less than this
               sigma_max_change=0.2,         # clips adaptive sigma to 20%
               learning_rate=0.01,           # learning rate for standard deviation
               learning_rate_decay = 0.9999, # annealing the learning rate
               learning_rate_limit = 0.01,   # stop annealing learning rate
               elite_ratio = 0,              # if > 0, then ignore learning_rate
               popsize=256,                  # population size
               average_baseline=True,        # set baseline to average of batch
               weight_decay=0.01,            # weight decay coefficient
               rank_fitness=True,            # use rank rather than fitness numbers
               forget_best=True):            # don't keep the historical best solution

    self.num_params = num_params
    self.sigma_init = sigma_init
    self.sigma_alpha = sigma_alpha
    self.sigma_decay = sigma_decay
    self.sigma_limit = sigma_limit
    self.sigma_max_change = sigma_max_change
    self.learning_rate = learning_rate
    self.learning_rate_decay = learning_rate_decay
    self.learning_rate_limit = learning_rate_limit
    self.popsize = popsize
    self.average_baseline = average_baseline
    if self.average_baseline:
      assert (self.popsize % 2 == 0), "Population size must be even"
      self.batch_size = int(self.popsize / 2)
    else:
      assert (self.popsize & 1), "Population size must be odd"
      self.batch_size = int((self.popsize - 1) / 2)

    # option to use greedy es method to select next mu, rather than using drift param
    self.elite_ratio = elite_ratio
    self.elite_popsize = int(self.popsize * self.elite_ratio)
    self.use_elite = False
    if self.elite_popsize > 0:
      self.use_elite = True

    self.forget_best = forget_best
    self.batch_reward = np.zeros(self.batch_size * 2)
    self.mu = np.zeros(self.num_params)
    self.sigma = np.ones(self.num_params) * self.sigma_init
    self.curr_best_mu = np.zeros(self.num_params)
    self.best_mu = np.zeros(self.num_params)
    self.best_reward = 0
    self.first_interation = True
    self.weight_decay = weight_decay
    self.rank_fitness = rank_fitness
    if self.rank_fitness:
      self.forget_best = True # always forget the best one if we rank
    # choose optimizer
    self.optimizer = Adam(self, learning_rate)

  def rms_stdev(self):
    sigma = self.sigma
    return np.mean(np.sqrt(sigma*sigma))

  def ask(self):
    '''returns a list of parameters'''
    # antithetic sampling
    self.epsilon = np.random.randn(self.batch_size, self.num_params) * self.sigma.reshape(1, self.num_params)
    self.epsilon_full = np.concatenate([self.epsilon, - self.epsilon])
    if self.average_baseline:
      epsilon = self.epsilon_full
    else:
      # first population is mu, then positive epsilon, then negative epsilon
      epsilon = np.concatenate([np.zeros((1, self.num_params)), self.epsilon_full])
    solutions = self.mu.reshape(1, self.num_params) + epsilon
    self.solutions = solutions
    return solutions

  def tell(self, reward_table_result):
    # input must be a numpy float array
    assert(len(reward_table_result) == self.popsize), "Inconsistent reward_table size reported."

    reward_table = np.array(reward_table_result)

    if self.rank_fitness:
      reward_table = compute_centered_ranks(reward_table)

    if self.weight_decay > 0:
      l2_decay = compute_weight_decay(self.weight_decay, self.solutions)
      reward_table += l2_decay

    reward_offset = 1
    if self.average_baseline:
      b = np.mean(reward_table)
      reward_offset = 0
    else:
      b = reward_table[0] # baseline

    reward = reward_table[reward_offset:]
    if self.use_elite:
      idx = np.argsort(reward)[::-1][0:self.elite_popsize]
    else:
      idx = np.argsort(reward)[::-1]

    best_reward = reward[idx[0]]
    if (best_reward > b or self.average_baseline):
      best_mu = self.mu + self.epsilon_full[idx[0]]
      best_reward = reward[idx[0]]
    else:
      best_mu = self.mu
      best_reward = b

    self.curr_best_reward = best_reward
    self.curr_best_mu = best_mu

    if self.first_interation:
      self.sigma = np.ones(self.num_params) * self.sigma_init
      self.first_interation = False
      self.best_reward = self.curr_best_reward
      self.best_mu = best_mu
    else:
      if self.forget_best or (self.curr_best_reward > self.best_reward):
        self.best_mu = best_mu
        self.best_reward = self.curr_best_reward

    # short hand
    epsilon = self.epsilon
    sigma = self.sigma

    # update the mean

    # move mean to the average of the best idx means
    if self.use_elite:
      self.mu += self.epsilon_full[idx].mean(axis=0)
    else:
      rT = (reward[:self.batch_size] - reward[self.batch_size:])
      change_mu = np.dot(rT, epsilon)
      self.optimizer.stepsize = self.learning_rate
      update_ratio = self.optimizer.update(-change_mu) # adam, rmsprop, momentum, etc.
      #self.mu += (change_mu * self.learning_rate) # normal SGD method

    # adaptive sigma
    # normalization
    if (self.sigma_alpha > 0):
      stdev_reward = 1.0
      if not self.rank_fitness:
        stdev_reward = reward.std()
      S = ((epsilon * epsilon - (sigma * sigma).reshape(1, self.num_params)) / sigma.reshape(1, self.num_params))
      reward_avg = (reward[:self.batch_size] + reward[self.batch_size:]) / 2.0
      rS = reward_avg - b
      delta_sigma = (np.dot(rS, S)) / (2 * self.batch_size * stdev_reward)

      # adjust sigma according to the adaptive sigma calculation
      # for stability, don't let sigma move more than 10% of orig value
      change_sigma = self.sigma_alpha * delta_sigma
      change_sigma = np.minimum(change_sigma, self.sigma_max_change * self.sigma)
      change_sigma = np.maximum(change_sigma, - self.sigma_max_change * self.sigma)
      self.sigma += change_sigma

    if (self.sigma_decay < 1):
      self.sigma[self.sigma > self.sigma_limit] *= self.sigma_decay

    if (self.learning_rate_decay < 1 and self.learning_rate > self.learning_rate_limit):
      self.learning_rate *= self.learning_rate_decay

  def current_param(self):
    return self.curr_best_mu

  def set_mu(self, mu):
    self.mu = np.array(mu)

  def best_param(self):
    return self.best_mu

  def result(self): # return best params so far, along with historically best reward, curr reward, sigma
    return (self.best_mu, self.best_reward, self.curr_best_reward, self.sigma)



import numpy as np


class Opt():
    """Base class for minimizer.

    Note:
        Concrete minimizers should extend this class.

    """
    name = 'Base minimizer'

    def __init__(self, func, d, a, b, x_min=None, y_min=None, verb=True):
        self._f0 = func          # Scalar function
        self._f = func           # Vector function
        self.d = d               # Dimension
        self.a = a               # Grid lower limit
        self.b = b               # Grid upper limit
        self.x_real = x_min      # Real x (arg) for minimum (for error check)
        self.y_real = y_min      # Real y (f(x)) for minimum (for error check)
        self.verb = verb         # Verbosity of output (True/False)

        self.prep()
        self.init()

    @property
    def e_x(self):
        if self.x_real is None or self.x is None:
            return None
        return np.linalg.norm(self.x_real - self.x)

    @property
    def e_y(self):
        if self.y_real is None or self.y is None:
            return None
        return abs(self.y_real - self.y)

    def info(self, text_spec=''):
        text = ''
        text += f'{self.name}' + ' '*(12 - len(self.name)) + ' | '

        text += f't={self.t:-6.1f}'

        if self.e_y is not None:
            text += f' | ey={self.e_y:-7.1e}'

        text += f' | evals={self.m:-7.1e}'

        if text_spec:
            text += f' | {text_spec}'

        return text

    def init(self):
        self.t = 0.              # Work time (sec)
        self.m = 0               # Number of function calls
        self.x = None            # Found x (arg) for minimum
        self.y = None            # Found y (f(x)) for minimum

        return self

    def f0(self, x):
        self.m += 1
        return self._f0(x)

    def f0_max(self, x):
        return -self.f0(x)

    def f(self, X):
        self.m += X.shape[0]
        return self._f(X)

    def f_max(self, X):
        return -self.f(X)

    def prep(self):
        return self

    def run_estool(self, solver):
        for j in range(self.iters):
            solutions = solver.ask()

            fitness_list = np.zeros(solver.popsize)
            for i in range(solver.popsize):
                fitness_list[i] = self.f_max(solutions[i].reshape(1, -1))[0]

            solver.tell(fitness_list)
            result = solver.result()

            self.x = result[0]
            self.y = result[1]

            if self.verb and (j+1) % 10 == 0:
                text = ''
                text += f'k={self.m:-8.2e} | '
                text += f'iter={j+1:-6d} | '
                text += f'e_y={self.e_y:-8.2e} '
                print(text)

    def solve(self):
        raise NotImplementedError()

    def to_dict(self):
        return {
            'name': self.name,
            'd': self.d,
            'a': self.a,
            'b': self.b,
            't': self.t,
            'm': self.m,
            'y': self.y,
            'y_real': self.y_real,
            'e_x': self.e_x,
            'e_y': self.e_y,
        }



from time import perf_counter as tpc
from ttopt import TTOpt


from opt import Opt


class OptTTOpt(Opt):
    """Minimizer based on the TTOpt."""
    name = 'TTOpt'

    def info(self):
        if self.n is not None:
            text = f'n={self.n:-5d}'
        else:
            text = f'p={self.p:-1d}, q={self.q:-2d}'
        text += f', r={self.r:-3d}'
        return super().info(text)

    def prep(self, n=None, p=2, q=12, r=6, evals=1.E+7):
        self.n = n
        self.p = p
        self.q = q
        self.r = r
        self.evals = int(evals)

        return self

    def solve(self):
        t = tpc()

        ttopt = TTOpt(
            self.f,
            d=self.d,
            a=self.a,
            b=self.b,
            n=self.n,
            p=self.p,
            q=self.q,
            evals=self.evals,
            x_min_real=self.x_real,
            y_min_real=self.y_real,
            with_log=self.verb)
        ttopt.minimize(self.r)

        self.t = tpc() - t

        self.x = ttopt.x_min
        self.y = ttopt.y_min

    def to_dict(self):
        res = super().to_dict()
        res['n'] = self.n
        res['p'] = self.p
        res['q'] = self.q
        res['r'] = self.r
        res['evals'] = self.evals
        return res



"""The demo of using ttopt. Example with QTT.

We'll find the minimum for the 10-dimensional Alpine function with vectorized
input. The target function for minimization has the form f(X), where input X is
the [samples, dimension] numpy array.

The target function and all the selected parameters are the same as in the
"base.py", except that we replace grid size "n" by grid-factors "p" and "q".

Run it from the root of the project as "python demo/qtt.py".

As a result of the script work we expect the output in console like this:
"
...
Alpine-10d | k=1.00e+05 | t_cur=1.66e-01 | e_x=3.32e+00 e_y=6.22e-06
----------------------------------------------------------------------
Alpine-10d | k=1.00e+05 | t_all=2.56e+00 | e_x=3.32e+00 e_y=6.22e-06
"

"""
import numpy as np


from ttopt import TTOpt
from ttopt import ttopt_init


np.random.seed(42)


d = 10                      # Number of function dimensions:
rmax = 4                    # Maximum TT-rank while cross-like iterations
def f(X):                   # Target function
    return np.sum(np.abs(X * np.sin(X) + 0.1 * X), axis=1)


# We initialize the TTOpt class instance with the correct parameters:
tto = TTOpt(
    f=f,                    # Function for minimization. X is [samples, dim]
    d=d,                    # Number of function dimensions
    a=-10.,                 # Grid lower bound (number or list of len d)
    b=+10.,                 # Grid upper bound (number or list of len d)
    p=2,                    # The grid size factor (there will n=p^q points)
    q=20,                   # The grid size factor (there will n=p^q points)
    evals=1.E+5,            # Number of function evaluations
    name='Alpine',          # Function name for log (this is optional)
    x_min_real=np.ones(d),  # Real value of x-minima (x; this is for test)
    y_min_real=0.,          # Real value of y-minima (y=f(x); this is for test)
    with_log=True)


# And now we launching the minimizer:
tto.minimize(rmax)


# We can extract the results of the computation:
x = tto.x_min          # The found value of the minimum of the function (x)
y = tto.y_min          # The found value of the minimum of the function (y=f(x))
k_c = tto.k_cache      # Total number of cache usage (should be 0 in this demo)
k_e = tto.k_evals      # Total number of requests to func (is always = evals)
k_t = tto.k_total      # Total number of requests (k_cache + k_evals)
t_f = tto.t_evals_mean # Average time spent to real function call for 1 point
                       # ... (see "ttopt.py" and docs for more details)


# We log the final state:
print('-' * 70 + '\n' + tto.info() +'\n\n')



from time import perf_counter as tpc


from es import OpenES
from opt import Opt


class OptES(Opt):
    """Minimizer based on the OpenAI-ES."""
    name = 'ES-OpenAI'

    def prep(self, popsize=256, iters=40000, sigma=0.1, decay=0.01):
        self.popsize = int(popsize)
        self.iters = int(iters)
        self.sigma = sigma
        self.decay = decay

        return self

    def solve(self):
        t = tpc()

        self.run_estool(OpenES(
            self.d,
            popsize=self.popsize,
            rank_fitness=False,
            sigma_init=self.sigma,
            weight_decay=self.decay))

        self.t = tpc() - t

    def to_dict(self):
        res = super().to_dict()
        res['popsize'] = self.popsize
        res['iters'] = self.iters
        res['sigma'] = self.sigma
        res['decay'] = self.decay
        return res



from time import perf_counter as tpc
import nevergrad as ng


from opt import Opt


class OptPSO(Opt):
    """Minimizer based on the PSO method from "nevergrad" package."""
    name = 'PSO'

    def prep(self, evals=1.E+7):
        self.evals = int(evals)

        return self

    def solve(self):
        t = tpc()

        par = ng.p.Array(shape=(self.d,), lower=self.a, upper=self.b)
        opt = ng.optimizers.registry['PSO'](budget=self.evals,
            parametrization=par, num_workers=1)
        self.x = opt.minimize(self.f0).value
        self.y = self.f0(self.x)

        self.t = tpc() - t


    def to_dict(self):
        res = super().to_dict()
        res['evals'] = self.evals
        return res



"""Investigating TTOpt performance for analytical benchmark functions."""
import argparse
import numpy as np
import pickle
import random
import sys


import matplotlib.pyplot as plt
params = {
    'text.usetex' : False,
    'font.size' : 36,
    'legend.fancybox':True,
    'legend.loc' : 'best',
    'legend.framealpha': 0.9,
    "legend.fontsize" : 27}
plt.rcParams.update(params)


sys.path.append('./demo_calc/opt')
from opt_tt_opt import OptTTOpt
from opt_ga import OptGA
from opt_es import OptES
from opt_es_cma import OptESCMA
from opt_de import OptDE
from opt_nb import OptNB
from opt_pso import OptPSO


from teneva import func_demo_all


# Minimizer classes:
OPTS = [OptTTOpt, OptGA, OptES, OptESCMA, OptDE, OptNB, OptPSO]


# Minimizer names:
OPT_NAMES = {
    'TTOpt': 'TTOpt',
    'GA': 'GA',
    'ES-OpenAI': 'openES',
    'ES-CMA': 'cmaES',
    'DE': 'DE',
    'NB': 'NB',
    'PSO': 'PSO',
}


# Function names and possible dimensions (set "True" if works for any):
FUNCS = {
    'Ackley': True,
    'Alpine': True,
    'Brown': True,
    'Exponential': True,
    'Grienwank': True,
    'Michalewicz': [10],
    'Qing': True,
    'Rastrigin': True,
    'Schaffer': True,
    'Schwefel': True,
}


# List of dimensions to check the TTOpt for multi-dim case:
D_LIST = [10, 50, 100, 500]

# List of ranks to check dependency of TTOpt on rank:
R_LIST = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]


# List of grid sizes to check QTT-effect (should be power of 2 and 4):
N_LIST = [2**8, 2**10, 2**12, 2**14, 2**16, 2**18, 2**20]


# List of numbers of function calls to check the dependency:
EVALS_LIST = [1.E+4, 5.E+4, 1.E+5, 5.E+5, 1.E+6, 5.E+6, 1.E+7]


# Population size for genetic-based algorithms:
GA_POPSIZE = 255


def get_funcs(d):
    names = []
    for name, dims in FUNCS.items():
        if isinstance(dims, list):
            if not d in dims:
                continue
        elif not dims:
            continue
        names.append(name)
    return func_demo_all(d, names=names)


def get_opt(func, opt_class, n, p, q, r, evals, with_log=False):
    opt = opt_class(func.get_f_poi, func.d, func.a, func.b,
        func.x_min, func.y_min, verb=with_log)

    if opt.name == 'TTOpt':
        opt.prep(n, p, q, r, evals)
    elif opt.name in ['DE', 'NB', 'PSO']:
        opt.prep(evals)
    else:
        opt.prep(popsize=GA_POPSIZE, iters=evals/GA_POPSIZE)

    return opt


def load(d, name, kind):
    fpath = f'./demo_calc/res_data/{name}_{kind}_{d}dim.pickle'
    try:
        with open(fpath, 'rb') as f:
            res = pickle.load(f)
    except Exception as e:
        res = None

    return res


def log(text, d, name, kind, is_init=False):
    print(text)

    fpath = f'./demo_calc/res_logs/{name}_{kind}_{d}dim.txt'
    with open(fpath, 'w' if is_init else 'a') as f:
        f.write(text + '\n')


def run_comp(d, p, q, r, evals, reps=1, name='calc1', with_log=False):
    """Compare different methods for benchmark analytic functions."""
    log(f'', d, name, 'comp', is_init=True)
    res = {}

    for func in get_funcs(d):
        log(f'--- Minimize function {func.name}-{d}dim\n', d, name, 'comp')
        res[func.name] = {}

        for opt_class in OPTS:
            opt = get_opt(func, opt_class, None, p, q, r, evals, with_log)
            res[func.name][opt.name] = solve(opt, d, name, 'comp', reps)

            save(res, d, name, 'comp')

        log('\n\n', d, name, 'comp')


def run_dims(p, q, r, reps=1, name='calc1', with_log=False, evals_par=1.E+4):
    """Solve for different dimension numbers."""
    d0 = D_LIST[-1]

    log(f'', d0, name, 'dims', is_init=True)
    res = {}

    for d in D_LIST:
        evals = int(evals_par * d)
        res[d] = {}

        for func in get_funcs(d):
            log(f'--- Minimize function {func.name}-{d}dim', d0, name, 'dims')

            opt = get_opt(func, OptTTOpt, None, p, q, r, evals, with_log)

            res[d][func.name] = solve(opt, d, name, 'dims', reps, d0)

            save(res, d0, name, 'dims')

            log('', d0, name, 'dims')


def run_iter(d, p, q, r, reps=1, name='calc1', with_log=False):
    """Check dependency of TTOpt on evals for benchmark analytic functions."""
    log(f'', d, name, 'iter', is_init=True)
    res = {}

    for func in get_funcs(d):
        log(f'--- Minimize function {func.name}-{d}dim\n', d, name, 'iter')
        res[func.name] = []

        for evals in EVALS_LIST:
            opt = get_opt(func, OptTTOpt, None, p, q, r, evals, with_log)
            res[func.name].append(solve(opt, d, name, 'iter', reps))

            save(res, d, name, 'iter')

        log('\n\n', d, name, 'iter')


def run_quan(d, r, evals, reps=1, name='calc1', with_log=False):
    """Check effect of QTT-based approach for benchmark analytic functions."""
    log(f'', d, name, 'quan', is_init=True)
    res = {}

    for func in get_funcs(d):
        log(f'--- Minimize function {func.name}-{d}dim\n', d, name, 'quan')

        res[func.name] = {'q0': [], 'q2': [], 'q4': []}

        for n in N_LIST:
            n = int(n)
            q2 = int(np.log2(n))
            q4 = int(q2 / 2)

            if 2**q2 != n or 4**q4 != n:
                raise ValueError(f'Invalid grid size "{n}"')

            opt = get_opt(func, OptTTOpt, n, None, None, r, evals, with_log)
            res[func.name]['q0'].append(solve(opt, d, name, 'quan', reps))

            opt = get_opt(func, OptTTOpt, None, 2, q2, r, evals, with_log)
            res[func.name]['q2'].append(solve(opt, d, name, 'quan', reps))

            opt = get_opt(func, OptTTOpt, None, 4, q4, r, evals, with_log)
            res[func.name]['q4'].append(solve(opt, d, name, 'quan', reps))

            save(res, d, name, 'quan')

        log('\n\n', d, name, 'quan')


def run_rank(d, p, q, evals, reps=1, name='calc1', with_log=False):
    """Check dependency of TTOpt on rank for benchmark analytic functions."""
    log(f'', d, name, 'rank', is_init=True)
    res = {}
    for func in get_funcs(d):
        log(f'--- Minimize function {func.name}-{d}dim\n', d, name, 'rank')
        res[func.name] = []

        for r in R_LIST:
            opt = get_opt(func, OptTTOpt, None, p, q, r, evals, with_log)
            res[func.name].append(solve(opt, d, name, 'rank', reps))

            save(res, d, name, 'rank')

        log('\n\n', d, name, 'rank')


def run_show(d, name='calc1'):
    """Show results of the previous calculations."""
    log(f'', d, name, 'show', is_init=True)
    run_show_comp(d, name)
    run_show_dims(d, D_LIST[-1], name)
    run_show_iter(d, name)
    run_show_quan(d, name)
    run_show_rank(d, name)


def run_show_comp(d, name='calc1'):
    """Show results of the previous calculations for "comp"."""
    res = load(d, name, 'comp')
    if res is None:
        log('>>> Comp-result is not available\n\n', d, name, 'show')
        return

    text = '>>> Comp-result (part of latex table): \n\n'

    text += '% ------ AUTO CODE START\n\n'

    for name_opt, name_opt_text in OPT_NAMES.items():
        text += '\\multirow{2}{*}{\\func{' + name_opt_text + '}}'

        text += '\n& $\\epsilon$ '
        for func in get_funcs(d):
            v = res[func.name][name_opt]['e']
            vals = [res[func.name][nm]['e']
                for nm in OPT_NAMES.keys() if nm != name_opt]
            if v <= np.min(vals):
                text += '& \\textbf{' + f'{v:-8.1e}' + '} '
            else:
                text += f'& {v:-8.1e} '

        text += '\n\\\\'

        text += '\n& $\\tau$ '
        for func in get_funcs(d):
            v = res[func.name][name_opt]['t']
            text += '& \\textit{' + f'{v:-6.2f}' + '} '

        text += ' \\\\ \\hline \n\n'

    text += '% ------ AUTO CODE END\n\n'

    log(text, d, name, 'show')


def run_show_dims(d0, d_max, name='calc1'):
    res = load(d_max, name, 'dims')
    if res is None:
        log('>>> Dims-result is not available', d0, name, 'show')
        return

    text = '>>> Dims-result: \n\n'

    text += '% ------ AUTO CODE START\n'

    for i, name_func in enumerate(res[d_max].keys(), 1):
        text += '\\multirow{2}{*}{\\emph{F' + str(i) + '}}\n'

        text += '& $\\epsilon$ \n'
        for d in D_LIST:
            item = res[d][name_func]
            e = item['e']
            text += '& ' + f'{e:-8.1e}' + '\n'

        text += '\\\\ \n'

        text += '& $\\tau$ \n'
        for d in D_LIST:
            item = res[d][name_func]
            t = item['t']
            text += '& \\textit{' + f'{t:-.1f}' + '}\n'

        text += '\\\\ \\hline \n\n'

    text += '% ------ AUTO CODE END\n\n'

    log(text, d0, name, 'show')


def run_show_iter(d, name='calc1'):
    """Show results of the previous calculations for "iter"."""
    res = load(d, name, 'iter')
    if res is None:
        log('>>> Iter-result is not available', d, name, 'show')
        return

    text = '>>> Iter-result (png file with plot): \n\n'

    plt.figure(figsize=(16, 8))
    plt.xlabel('number of queries')
    plt.ylabel('absolute error')

    for i, func in enumerate(get_funcs(d), 1):
        v = [item['e'] for item in res[func.name]]
        plt.plot(EVALS_LIST, v, label=f'F{i}', marker='o')

    plt.grid()
    plt.semilogx()
    plt.semilogy()
    plt.legend(loc='best', ncol=5, fontsize=20)

    fpath = f'./demo_calc/res_plot/{name}_iter_{d}dim.png'
    plt.savefig(fpath, bbox_inches='tight')
    text += f'Figure saved to file "{fpath}"\n\n'

    log(text, d, name, 'show')


def run_show_quan(d, name='calc1'):
    """Show results of the previous calculations for "quan"."""
    res = load(d, name, 'quan')
    if res is None:
        log('>>> Quan-result is not available', d, name, 'show')
        return

    text = '>>> Quan-result (part of latex table): \n\n'

    text += '% ------ AUTO CODE START\n'

    for i, n in enumerate(N_LIST):
        text += '\\multirow{2}{*}{' + str(n) + '}'

        text += '\n& TT '
        for func in get_funcs(d):
            v = res[func.name]['q0'][i]['e']
            text += f'& {v:-8.1e} '
        text += '\\\\'

        text += '\n& QTT '
        for func in get_funcs(d):
            v = res[func.name]['q2'][i]['e']
            text += f'& {v:-8.1e} '
        text += ' \\\\ \\hline \n'

    text += '% ------ AUTO CODE END\n\n'

    log(text, d, name, 'show')


def run_show_rank(d, name='calc1'):
    """Show results of the previous calculations for "rank"."""
    res = load(d, name, 'rank')
    if res is None:
        log('>>> Rank-result is not available', d, name, 'show')
        return

    text = '>>> Rank-result (png file with plot): \n\n'

    plt.figure(figsize=(16, 8))
    plt.xlabel('rank')
    plt.ylabel('absolute error')
    plt.xticks(R_LIST)

    for i, func in enumerate(get_funcs(d), 1):
        v = [item['e'] for item in res[func.name]]
        plt.plot(R_LIST, v, label=f'F{i}', marker='o')

    plt.grid()
    plt.semilogy()
    plt.legend(loc='best', ncol=5, fontsize=20)

    fpath = f'./demo_calc/res_plot/{name}_rank_{d}dim.png'
    plt.savefig(fpath, bbox_inches='tight')
    text += f'Figure saved to file "{fpath}"\n\n'

    log(text, d, name, 'show')


def save(res, d, name, kind):
    fpath = f'./demo_calc/res_data/{name}_{kind}_{d}dim.pickle'
    with open(fpath, 'wb') as f:
        pickle.dump(res, f, protocol=pickle.HIGHEST_PROTOCOL)


def solve(opt, d, name, kind, reps=1, d_log=None):
    t, y, e, m = [], [], [], []
    for rep in range(reps):
        np.random.seed(rep)
        random.seed(rep)

        opt.init()
        opt.solve()

        log(opt.info(), d_log or d, name, kind)

        t.append(opt.t)
        y.append(opt.y)
        e.append(opt.e_y)
        m.append(opt.m)

    if reps > 1:
        print(f'--> Mean error / time : {np.mean(e):-7.1e} / {np.mean(t):.2f}')

    return {
        't': np.mean(t),
        'e': np.mean(e),
        'e_var': np.var(e),
        'e_min': np.min(e),
        'e_max': np.max(e),
        'e_avg': np.mean(e),
        'e_all': e,
        'y_all': y,
        'y_real': opt.y_real,
        'evals': int(np.mean(m))}


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Investigating TTOpt performance for analytic functions')
    parser.add_argument('-d', default=10,
        type=int, help='dimension d')
    parser.add_argument('-p', default=2,
        type=int, help='grid param p')
    parser.add_argument('-q', default=25,
        type=int, help='grid param q')
    parser.add_argument('-r', default=4,
        type=int, help='rank')
    parser.add_argument('--evals', default=1.E+5,
        type=float, help='computational budget')
    parser.add_argument('--reps', default=1,
        type=int, help='repetitions')
    parser.add_argument('--name', default='calc1',
        type=str, help='calculation name (the corresponding prefix will be used for the files with results)')
    parser.add_argument('--kind', default='comp',
        type=str, help='kind of calculation ("comp" - compare different solvers; "dims" - check dependency on dimension number; "iter" - check dependency on number of calls for target function; "quan" - check effect of qtt-usage; "rank" - check dependency on rank; "show" - show results of the previous calculations)')
    parser.add_argument('--verb', default=False,
        type=bool, help='if True, then intermediate results of the optimization process will be printed to the console')

    args = parser.parse_args()

    if args.kind == 'comp':
        run_comp(args.d, args.p, args.q, args.r, args.evals, args.reps,
            args.name, args.verb)
    elif args.kind == 'dims':
        run_dims(args.p, args.q, args.r, args.reps,
            args.name, args.verb)
    elif args.kind == 'iter':
        run_iter(args.d, args.p, args.q, args.r, args.reps,
            args.name, args.verb)
    elif args.kind == 'quan':
        run_quan(args.d, args.r, args.evals, args.reps,
            args.name, args.verb)
    elif args.kind == 'rank':
        run_rank(args.d, args.p, args.q, args.evals, args.reps,
            args.name, args.verb)
    elif args.kind == 'show':
        run_show(args.d, args.name)
    else:
        raise ValueError(f'Invalid kind of calculation "{args.kind}"')
